"""Module containing event bus implementation & handler description."""
import dataclasses
from typing import Any, Type

from yaeb.base.bus import BaseEventBus
from yaeb.base.events import AllEvents, BaseEvent
from yaeb.base.handlers import BaseEventHandler
from yaeb.base.registry import BaseEventHandlerRegistry


@dataclasses.dataclass
class DictEventHandlerRegistry(BaseEventHandlerRegistry):
    """Registry storing event handlers in memory."""

    event_handlers: dict[
        Type[BaseEvent] | Type[AllEvents],
        list[BaseEventHandler[Any]],
    ] = dataclasses.field(default_factory=dict)

    def add_event_handler(
        self,
        event_type: Type[BaseEvent] | Type[AllEvents],
        event_handler: BaseEventHandler[Any],
    ) -> None:
        """Add event handler for given event type to in memory storage."""
        if event_type not in self.event_handlers:
            self.event_handlers[event_type] = []

        self.event_handlers[event_type].append(event_handler)

    def get_event_handlers(
        self,
        event_type: Type[BaseEvent] | Type[AllEvents],
    ) -> tuple[BaseEventHandler[Any], ...]:
        """Get stored in memory event handlers for given event type."""
        return tuple(self.event_handlers.get(event_type, []))


@dataclasses.dataclass
class EventBus(BaseEventBus):
    """Implementation of the event bus."""

    event_handler_registry: BaseEventHandlerRegistry

    def register(
        self,
        event_type: Type[BaseEvent] | Type[AllEvents],
        event_handler: BaseEventHandler[Any],
    ) -> None:
        """Register handler for given event."""
        self.event_handler_registry.add_event_handler(event_type, event_handler)

    def emit(self, event: BaseEvent) -> None:
        """Send event to bus executing all registered handlers."""
        all_events_handlers = self.event_handler_registry.get_event_handlers(AllEvents)

        for all_events_handler in all_events_handlers:
            all_events_handler.execute(event=event, bus=self)

        event_handlers = self.event_handler_registry.get_event_handlers(type(event))

        for event_handler in event_handlers:
            event_handler.execute(event=event, bus=self)



"""Module storing base event handler registry class."""
from __future__ import annotations

import abc
from typing import TYPE_CHECKING, Any, Type

if TYPE_CHECKING:
    from yaeb.base.events import AllEvents, BaseEvent
    from yaeb.base.handlers import BaseEventHandler


class BaseEventHandlerRegistry(abc.ABC):
    """Registry handling adding/getting event handlers using any storage."""

    @abc.abstractmethod
    def add_event_handler(
        self,
        event_type: Type[BaseEvent] | Type[AllEvents],
        event_handler: BaseEventHandler[Any],
    ) -> None:
        """Add event handler for given event type to registry."""
        raise NotImplementedError

    @abc.abstractmethod
    def get_event_handlers(
        self,
        event_type: Type[BaseEvent] | Type[AllEvents],
    ) -> tuple[BaseEventHandler[Any], ...]:
        """Get event handlers for given event type from registry."""
        raise NotImplementedError



"""Module storing base event bus."""
from __future__ import annotations

import abc
from typing import TYPE_CHECKING, Any, Type

if TYPE_CHECKING:
    from yaeb.base.events import AllEvents, BaseEvent
    from yaeb.base.handlers import BaseEventHandler


class BaseEventBus(abc.ABC):
    """Base event bus interface."""

    @abc.abstractmethod
    def register(
        self,
        event_type: Type[BaseEvent] | Type[AllEvents],
        event_handler: BaseEventHandler[Any],
    ) -> None:
        """Register handler for given event."""
        raise NotImplementedError

    @abc.abstractmethod
    def emit(self, event: BaseEvent) -> None:
        """Send event to bus executing all registered handlers."""
        raise NotImplementedError



"""Module storing base event models."""
from __future__ import annotations

import abc
from collections import deque
from typing import final


class BaseEvent(abc.ABC):
    """Base class for events."""

    parent_event: BaseEvent | None

    def __init__(self, parent_event: BaseEvent | None) -> None:
        """Add parent event for further tracking."""
        self.parent_event = parent_event

    def get_history(self) -> tuple[BaseEvent, ...]:
        """Get history of events resulted current event to be emitted."""
        history: deque[BaseEvent] = deque()

        parent_event = self.parent_event
        while parent_event is not None:
            history.appendleft(parent_event)
            parent_event = parent_event.parent_event

        return tuple(history)


@final
class AllEvents:
    """Event type allowing to subscribe to all events."""



"""Module storing base event bus handlers."""
from __future__ import annotations

import abc
import asyncio
from typing import TYPE_CHECKING, Generic, TypeVar

from yaeb.base.events import BaseEvent

if TYPE_CHECKING:
    from concurrent.futures import Executor

    from yaeb.base.bus import BaseEventBus

E = TypeVar('E', bound=BaseEvent)


class BaseEventHandler(abc.ABC, Generic[E]):
    """Base class for event handlers."""

    @abc.abstractmethod
    def execute(self, event: E, bus: BaseEventBus) -> None:
        """Execute handling of event."""
        raise NotImplementedError


class BaseAsyncEventHandler(BaseEventHandler[E]):
    """Base class for async event handlers."""

    @abc.abstractmethod
    async def handle_event(self, event: E, bus: BaseEventBus) -> None:
        """Handle given event asynchronously."""
        raise NotImplementedError

    def execute(self, event: E, bus: BaseEventBus) -> None:
        """Execute event handling via asyncio task asynchronously."""
        asyncio.Task(self.handle_event(event=event, bus=bus))


class BaseSyncEventHandler(BaseEventHandler[E]):
    """Base class for sync event handlers."""

    @abc.abstractmethod
    def handle_event(self, event: E, bus: BaseEventBus) -> None:
        """Handle given event synchronously."""
        raise NotImplementedError

    def execute(self, event: E, bus: BaseEventBus) -> None:
        """Execute handling of event simply by calling handle event method."""
        self.handle_event(event=event, bus=bus)


class BaseExecutorEventHandler(BaseEventHandler[E]):
    """Base event handler handling events via given executor."""

    executor: Executor

    def __init__(self, executor: Executor) -> None:
        """Add executor for further handle event execution."""
        self.executor = executor

    @abc.abstractmethod
    def handle_event(self, event: E, bus: BaseEventBus) -> None:
        """Handle event in the given executor."""
        raise NotImplementedError

    def execute(self, event: E, bus: BaseEventBus) -> None:
        """Pass handle event to given executor."""
        self.executor.submit(self.handle_event, event=event, bus=bus)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['yaeb', 'yaeb.base']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'yaeb',
    'version': '0.4.2',
    'description': 'A simple typed event bus written in pure python',
    'long_description': "\n# Yet another event bus - yaeb for short\n\nA simple typed event bus written in pure python\n\n\n## Installation\n\nInstall yaeb with pip\n\n```bash\n  pip install yaeb\n```\n    \n## Usage/Examples\n\n```python\nfrom logging import info\n\nfrom yaeb.base.bus import BaseEventBus\nfrom yaeb.base.events import BaseEvent\nfrom yaeb.base.handlers import BaseSyncEventHandler\nfrom yaeb.bus import DictEventHandlerRegistry, EventBus\n\n\nclass UserCreated(BaseEvent):\n    user_id: int\n\n    def __init__(self, user_id: int) -> None:\n        self.user_id = user_id\n\n\nclass UserCreatedHandler(BaseSyncEventHandler[UserCreated]):\n    def handle_event(self, event: UserCreated, bus: BaseEventBus) -> None:\n        info('User with id=%d was created!', event.user_id)\n\n\nif __name__ == '__main__':\n    bus = EventBus(event_handler_registry=DictEventHandlerRegistry())\n    bus.register(event_type=UserCreated, event_handler=UserCreatedHandler())\n\n    bus.emit(UserCreated(user_id=1))  # prints log message with created user id\n\n```\n\n\n## Roadmap\n\n- [x] Add coroutines support - Added `BaseAsyncEventHandler`\n- [x] Add some kind of multithreading support. Though it can be implemented by handlers themselves 🤔 - Added `BaseExecutorEventHandler`\n\n",
    'author': 'Daniil Fedyaev',
    'author_email': 'wintercitizen@outlook.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/WinterCitizen/yaeb',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
